"""
Meta Tuner Class: Used to optimize across a set of models:
- selecting intelligently the order of functions to optimize
Current implementation: Bare Metal functionality for testing.
ToDo: Improve code with better config management and remove hardcoded parameters
"""

from dataclasses import dataclass
from mango.domain.domain_space import DomainSpace
from mango.optimizer.bayesian_learning import BayesianLearning
import numpy as np
from tqdm.auto import tqdm
import random

import warnings

warnings.filterwarnings("ignore")


class MetaTuner:
    @dataclass
    class Config:
        n_iter: int = 20
        n_init: int = 2
        log_progress: bool = True

    def __init__(self, param_dict_list, objective_list, **kwargs):
        self.param_dict_list = param_dict_list
        self.objective_list = objective_list
        self.config = MetaTuner.Config(**kwargs)
        # adjustable parameters
        self.num_of_iterations = self.config.n_iter
        self.initial_random = self.config.n_init
        self.log_progress = self.config.log_progress

        # list of GPR for each objective
        self.gpr_list = []

        # store the results of MetaTuner
        self.results = dict()

        # batch size of the entire metaTuner
        self.batch_size = 1

        # batch size per function to select
        self.obj_batch_size = 1

        # use to see the info and fix the seeds metaTuner is running
        self.debug = False

        self.surrogate = None  # used to test different kernel functions

        # stores the index of evaluated objectives
        self.objectives_evaluated = []

        # stores the values of objectives in the order they are evaluated
        self.objective_values_list = []

        self.seed = 0

        # rate of decay of exploration probability
        self.decay_rate = 0.90

        # The initial exploration probability
        self.exploration_rate = 1.0
        self.exploration_rate_min = 0.1

        # record which classifier/gp was used last
        self.last_used = 0

        self.store_domain = {}
        self.first_itr = {}

    def run(self):
        # run the metaTuner
        self.results = self.runExponentialTuner()
        return self.results

    # return the max value of Y from current evaluations
    def get_max_y_value(self, Y_dict_array):
        values = Y_dict_array[0]

        for k in range(len(self.objective_list) - 1):
            values = np.vstack((values, Y_dict_array[k + 1]))

        max_val = np.max(values)

        return max_val

    def runExponentialTuner(self):
        """
        Steps:
        1-Create DS obj for each obj of param_dict_list
        2-Sample randomly from the DS objects and evaluate the objective functions.
        3- Now use the GPR for each objective to select next batch
        4- Select the best values based on fxn(surrogate) from GPR.
        5- Modify the surrogate selections so as to avoid getting struck.
        """

        results = dict()

        # random params tried
        results["random_params"] = []

        # random objective values
        results["random_params_objective"] = []

        # random objective function id
        results["random_objective_fid"] = []

        results["params_tried"] = []
        results["objective_values"] = []
        results["objective_fid"] = []

        # print('*** Entering Metatuner *** ')
        num_of_random = self.initial_random

        ds = []
        for i in self.param_dict_list:
            ds.append(DomainSpace(i))

        # dict of list for each obj function
        X_dict_list = {}
        Y_dict_list = {}

        # dict of array for each obj function
        X_dict_array = {}
        Y_dict_array = {}

        # stores the maximum value of the objective function for each objective
        Y_dict_array_max = {}

        # randomly evaluate the initial points for objectives
        for i in range(len(self.param_dict_list)):

            # used later
            self.first_itr[i] = True

            if self.debug:
                # fixing random seeds to reproduce the results for debugging
                np.random.seed(self.seed)
                random.seed(self.seed)

            ds_i = ds[i]

            # make sure not all values in y are same
            done = False

            # count no of times random repeated
            num_times = 0
            while not done:
                random_hyper_parameters = ds_i.get_random_sample(num_of_random)
                y_list = self.objective_list[i](random_hyper_parameters)
                # print(i, random_hyper_parameters, y_list)
                num_times = num_times + 1

                # 10 times done still there is a tie
                if num_times == 10:
                    done = True
                    y_list[0] = y_list[0] - 0.001

                for ind in range(num_of_random):
                    if y_list[0] != y_list[ind]:
                        done = True

            # print('Random_hyper_parameters:',random_hyper_parameters, y_list)
            x_list = random_hyper_parameters

            # print(i, random_hyper_parameters, x_list, y_list)

            X_dict_list[i] = []
            X_dict_list[i].append(x_list)

            Y_dict_list[i] = []
            Y_dict_list[i].append(y_list)

            self.objective_values_list += y_list

            results["random_params"] = results["random_params"] + x_list
            results["random_params_objective"] = (
                results["random_params_objective"] + y_list
            )
            results["random_objective_fid"] = results["random_objective_fid"] + [
                i
            ] * len(x_list)

            # x_array2 = ds_i.convert_GP_space(random_hyper_parameters)
            # x_array = ds_i.convert_to_gp(random_hyper_parameters)

            x_array = ds_i.convert_GP_space(random_hyper_parameters)

            X_dict_array[i] = x_array

            # print(x_array2)
            # print(x_array)

            y_array = np.array(y_list).reshape(len(y_list), 1)
            Y_dict_array[i] = y_array

            # the random ones are added as it is
            Y_dict_array_max[i] = y_array

        # Initialize the number of Optimizers
        Optimizer_list = []

        for i in range(len(self.objective_list)):
            Optimizer_i = BayesianLearning(surrogate=self.surrogate)
            Optimizer_i.domain_size = ds[i].domain_size
            Optimizer_list.append(Optimizer_i)

            # print('Domain Size:',ds[i].domain_size)

        # Storing optimizers attributes: iteration count and modified exploration of external surrogate
        Optimizer_exploration = []
        Optimizer_iteration = []
        for i in range(len(self.objective_list)):
            Optimizer_exploration.append(1.0)
            Optimizer_iteration.append(1.0)

        results["params_tried"] = results["random_params"]
        results["objective_values"] = results["random_params_objective"]
        results["objective_fid"] = results["random_objective_fid"]

        # print(Optimizer_exploration)

        # Now run the optimization iterations
        iterator = (
            tqdm(range(self.num_of_iterations))
            if self.log_progress
            else range(self.num_of_iterations)
        )

        for itr in iterator:
            # next values of x returned from individual function
            # Values in x are dependent on types of param dict, so using a list
            x_values_list = []

            # Next promising surrogate values for each objective: used to select functions
            s_values_array = np.empty((0, 1), float)

            # keeping track of objective indices: Maps the x values to its objective function
            x_obj_indices = []

            # Next promising surrogate values for each objective in list form
            s_values_list = []

            # sample individual domains and evaluate surrogate functions.
            # we get the next promising samples along with the surrogate function values
            max_val_y = self.get_max_y_value(Y_dict_array)

            # In GPs this value is used to scale the Y values
            max_val_y_scaled = max_val_y
            # In this setting the default GP performs reasonably well
            if max_val_y_scaled < 1.0:
                max_val_y_scaled = 1.0

            for j in range(len(ds)):
                if self.debug:
                    # fixing random seeds to reproduce the results for debugging
                    np.random.seed(self.seed)
                    random.seed(self.seed)

                # X_domain_np = ds[j].sample_gp_space()

                if self.first_itr[j]:
                    # if j==self.last_used or self.first_itr[j]:
                    # Domain Space
                    domain_list = ds[j].get_domain()
                    self.store_domain[j] = ds[j].convert_GP_space(domain_list)
                    # self.store_domain[j] = ds[j].sample_gp_space()
                    self.first_itr[j] = True
                    # print('doing domain sampling')

                X_domain_np = self.store_domain[j]

                # next batch of x for this objective along with its surrogate value
                X_next_batch, surr_value, surr_value_ext, u_values = Optimizer_list[
                    j
                ].get_next_batch_MetaTuner(
                    X_dict_array[j],
                    Y_dict_array[j] / max_val_y_scaled,
                    X_domain_np,
                    self.obj_batch_size,
                    Optimizer_exploration[j],
                    Optimizer_iteration[j],
                    classifier_index=j,
                    last_used_index=j,
                )  # self.last_used

                # used for displaying and debugging
                # s_values_list.append([int(surr_value_ext[0]*100.0),int(surr_value[0]*100.0)])
                s_values_list.append([int(surr_value_ext[0] * 100.0)])

                # this is used to sort the objective values
                s_value = np.array(surr_value_ext)  # modified sandeep

                # s_value = np.array(surr_value_ext)
                s_value = s_value.reshape(-1, s_value.shape[0])

                s_values_array = np.append(s_values_array, s_value)
                x_values_list = x_values_list + X_next_batch

                # keep track of objective function for corresponding surrogate and x values
                for k in range(self.obj_batch_size):
                    x_obj_indices.append(j)

            # sort the surrogate values in descending order, to select the best value from them
            v_sorting_index = np.argsort(-s_values_array, axis=0)

            # now select the self.batch_size values from x_values_list based on v_sorting_index
            v_sorting_index = v_sorting_index[: self.batch_size]

            # select randomly with exploration rate else select the max surrogate

            prob_selection = random.random()

            random_selected = False

            # do the random evaluation of the functions
            if prob_selection < self.exploration_rate:
                selected_obj = random.randint(
                    0, (len(ds) - 1)
                )  # 0 and len(ds included)

                if self.exploration_rate > self.exploration_rate_min:
                    self.exploration_rate = self.exploration_rate * self.decay_rate

                else:
                    self.exploration_rate = self.exploration_rate_min

                random_selected = True

            else:
                selected_obj = v_sorting_index[0]

            # keep track of objective indices selected in current iteration
            loc_indices = []
            curr_x_next_np = x_values_list[selected_obj]

            # convert this into the parameter space for scheduling
            # see the function index for this x value
            index = selected_obj

            # updating last used classifier
            self.last_used = index

            # keep track of local indices
            loc_indices.append(index)

            # keep track of indices in a global datastr for visualization of function selection
            self.objectives_evaluated.append(index)

            curr_x_next = ds[index].convert_PS_space(curr_x_next_np)

            # run the next curr_x_next value for the objective function
            y_list = self.objective_list[index](curr_x_next)

            self.objective_values_list += y_list

            results["params_tried"] = results["params_tried"] + curr_x_next
            results["objective_values"] = results["objective_values"] + y_list
            results["objective_fid"] = results["objective_fid"] + [index]

            curr_y_array = np.array(y_list).reshape(len(y_list), 1)
            # append the curr_x_next_np, curr_x_next, y_list to appropriate datastructures for book keeping

            X_dict_array[index] = np.vstack((X_dict_array[index], curr_x_next_np))
            Y_dict_array[index] = np.vstack((Y_dict_array[index], curr_y_array))

            Y_dict_array_max[index] = np.vstack(
                (Y_dict_array_max[index], np.max(Y_dict_array[index]))
            )

            # scale the exploration of objectives that are not selected for others make their exploration to 1
            for i in range(len(self.objective_list)):
                if i in loc_indices:
                    Optimizer_exploration[i] = 1.0
                    Optimizer_iteration[i] += 1.0
                else:
                    Optimizer_exploration[i] = (
                        Optimizer_exploration[i] + 1.1 * self.exploration_rate
                    )

            if self.log_progress:
                tqdm.write(": Best score: %s" % max_val_y)

            # print(itr, s_values_list, Optimizer_iteration, Optimizer_exploration, max_val_y)#, Y_dict_array[selected_obj])

            # print(itr, s_values_list, random_selected, self.exploration_rate, prob_selection, Optimizer_iteration , max_val_y)

            # print(itr, s_values_list, random_selected, Optimizer_iteration , int(self.exploration_rate*100.0), max_val_y)

            self.X_dict_array = X_dict_array
            self.Y_dict_array = Y_dict_array
            self.Y_dict_array_max = Y_dict_array_max
            self.ds = ds

        self.X_dict_array = X_dict_array
        self.Y_dict_array = Y_dict_array
        self.Y_dict_array_max = Y_dict_array_max
        self.ds = ds

        index_best = np.argmax(results["objective_values"])
        results["best_objective"] = results["objective_values"][index_best]
        results["best_params"] = results["params_tried"][index_best]
        results["best_objective_fid"] = results["objective_fid"][index_best]

        return results



"""
Main Tuner Class which uses other abstractions.
General usage is to find the optimal hyperparameters of the classifier
"""

import copy
from collections.abc import Mapping
from dataclasses import dataclass
import logging
import random
from typing import Callable
import warnings

from tqdm.auto import tqdm
import numpy as np

from mango.domain.domain_space import DomainSpace
from mango.optimizer.bayesian_learning import BayesianLearning
from mango.domain.parameter_sampler import parameter_sampler

# setting warnings to ignore for now
warnings.filterwarnings("ignore")

_logger = logging.getLogger(__name__)


@dataclass
class TunerConfig:
    domain_size: int = None
    initial_random: int = 2
    initial_custom: list = None
    num_iteration: int = 20
    batch_size: int = 1
    optimizer: str = "Bayesian"
    parallel_strategy: str = "clustering"
    surrogate: object = None  # used to test different kernel functions
    valid_optimizers = ["Bayesian", "Random"]
    valid_parallel_strategies = ["penalty", "clustering"]
    alpha: float = 2.0
    exploration: float = 1.0
    exploration_decay: float = 0.9
    exploration_min: float = 0.1
    fixed_domain: bool = False
    early_stopping: Callable = None
    constraint: Callable = None
    param_sampler: Callable = parameter_sampler
    scale_params: bool = False
    log_progress: bool = True

    def __post_init__(self):
        if self.optimizer not in self.valid_optimizers:
            raise ValueError(
                f"optimizer: {self.optimizer} is not valid, should be one of {self.valid_optimizers}"
            )
        if self.parallel_strategy not in self.valid_parallel_strategies:
            raise ValueError(
                f"parallel strategy: {self.parallel_strategy} is not valid,"
                f" should be one of {self.valid_parallel_strategies}"
            )

    @property
    def is_bayesian(self):
        return self.optimizer == "Bayesian"

    @property
    def is_random(self):
        return self.optimizer == "Random"

    @property
    def strategy_is_penalty(self):
        return self.parallel_strategy == "penalty"

    @property
    def strategy_is_clustering(self):
        return self.parallel_strategy == "clustering"

    def early_stop(self, results):
        if self.early_stopping is None:
            return False

        results = copy.deepcopy(results)
        return self.early_stopping(results)


class Tuner:

    def __init__(self, param_dict, objective, conf_dict=None):

        self.objective_function = objective
        self.maximize_objective = True

        if conf_dict is None:
            conf_dict = {}

        self.config = TunerConfig(**conf_dict)
        self.ds = DomainSpace(
            param_dict,
            param_sampler=self.config.param_sampler,
            constraint=self.config.constraint,
            scale_params=self.config.scale_params,
        )

        if self.config.domain_size is not None:
            self.ds.domain_size = self.config.domain_size

        # overwrite batch size if given as a property of objective function
        if hasattr(objective, "batch_size"):
            self.config.batch_size = objective.batch_size

        # stores the results of using the tuner
        self.results = dict()

    def run(self):
        if self.config.is_bayesian:
            self.results = self.runBayesianOptimizer()
        elif self.config.is_random:
            self.results = self.runRandomOptimizer()
        else:
            raise ValueError("Unknown Optimizer %s" % self.config.optimizer)

        return self.results

    def maximize(self):
        return self.run()

    def minimize(self):
        self.maximize_objective = False
        return self.run()

    def process_initial_custom(self):
        init_values = self.config.initial_custom

        if all(isinstance(v, Mapping) for v in init_values):
            X_init = copy.deepcopy(init_values)
            X_list, Y_list = self.runUserObjective(X_init)
            return X_list, Y_list
        elif all(isinstance(v, tuple) and len(v) == 2 for v in init_values):
            X_list = copy.deepcopy([v[0] for v in init_values])
            Y_list = np.array([v[1] for v in init_values])
            if self.maximize_objective is False:
                Y_list = -1 * Y_list
            return X_list, Y_list
        else:
            raise TypeError(
                f"Elements of initial_custom param should be either a dict of params or tuple (params, y),"
                f" got: {[type(v).__name__ for v in init_values]}"
            )

    def run_initial(self):
        if self.config.initial_custom is not None:
            return self.process_initial_custom()
        else:
            # getting first few random values
            X_init = self.ds.get_random_sample(self.config.initial_random)
            X_list, Y_list = self.runUserObjective(X_init)

            # in case initial random results are invalid try different samples
            n_tries = 1
            while len(Y_list) < self.config.initial_random and n_tries < 3:
                X_init = self.ds.get_random_sample(
                    self.config.initial_random - len(Y_list)
                )
                X_list2, Y_list2 = self.runUserObjective(X_init)
                X_list = np.append(X_list, X_list2)
                Y_list = np.append(Y_list, Y_list2)
                n_tries += 1

            if len(Y_list) == 0:
                raise ValueError(
                    "No valid configuration found to initiate the Bayesian Optimizer"
                )
        return X_list, Y_list

    def runBayesianOptimizer(self):
        results = dict()

        X_list, Y_list = self.run_initial()

        # evaluated hyperparameters are used
        X_init = self.ds.convert_GP_space(X_list)
        Y_init = Y_list.reshape(len(Y_list), 1)

        # setting the initial random hyperparameters tried
        results["random_params"] = X_list
        results["random_params_objective"] = Y_list

        Optimizer = BayesianLearning(
            surrogate=self.config.surrogate,
            alpha=self.config.alpha,
            domain_size=self.config.domain_size,
        )

        X_sample = X_init
        Y_sample = Y_init

        hyper_parameters_tried = X_list
        objective_function_values = Y_list
        surrogate_values = Y_list

        x_failed_evaluations = np.array([])

        domain_list = self.ds.get_domain()
        X_domain_np = self.ds.convert_GP_space(domain_list)

        # running the iterations
        iterator = (
            tqdm(range(self.config.num_iteration))
            if self.config.log_progress
            else range(self.config.num_iteration)
        )
        for _ in iterator:

            # adding a Minimum exploration to explore independent of UCB
            if random.random() < self.config.exploration:
                random_parameters = self.ds.get_random_sample(self.config.batch_size)
                X_next_batch = self.ds.convert_GP_space(random_parameters)

                if self.config.exploration > self.config.exploration_min:
                    self.config.exploration = (
                        self.config.exploration * self.config.exploration_decay
                    )

            elif self.config.strategy_is_penalty:
                X_next_batch = Optimizer.get_next_batch(
                    X_sample, Y_sample, X_domain_np, batch_size=self.config.batch_size
                )
            elif self.config.strategy_is_clustering:
                X_next_batch = Optimizer.get_next_batch_clustering(
                    X_sample, Y_sample, X_domain_np, batch_size=self.config.batch_size
                )
            else:
                # assume penalty approach
                X_next_batch = Optimizer.get_next_batch(
                    X_sample, Y_sample, X_domain_np, batch_size=self.config.batch_size
                )

            # Scheduler
            X_next_PS = self.ds.convert_PS_space(X_next_batch)

            # if all the xs have failed before, replace them with random sample
            # as we will not get any new information otherwise
            if all(x in x_failed_evaluations for x in X_next_PS):
                X_next_PS = self.ds.get_random_sample(self.config.batch_size)

            # Evaluate the Objective function
            X_next_list, Y_next_list = self.runUserObjective(X_next_PS)

            # keep track of all parameters that failed
            x_failed = [x for x in X_next_PS if x not in X_next_list]
            x_failed_evaluations = np.append(x_failed_evaluations, x_failed)

            if len(Y_next_list) == 0:
                # no values returned
                # this is problematic if domain is small and same value is tried again
                # in the next iteration as the optimizer would be stuck
                continue

            Y_next_batch = Y_next_list.reshape(len(Y_next_list), 1)
            # update X_next_batch to successfully evaluated values
            X_next_batch = self.ds.convert_GP_space(X_next_list)

            # update the bookeeping of values tried
            hyper_parameters_tried = np.append(hyper_parameters_tried, X_next_list)
            objective_function_values = np.append(
                objective_function_values, Y_next_list
            )
            surrogate_values = np.append(
                surrogate_values, Optimizer.surrogate.predict(X_next_batch)
            )

            # Appending to the current samples
            X_sample = np.vstack((X_sample, X_next_batch))
            Y_sample = np.vstack((Y_sample, Y_next_batch))

            # refresh domain if not fixed
            if not self.config.fixed_domain:
                domain_list = self.ds.get_domain()
                X_domain_np = self.ds.convert_GP_space(domain_list)

            results["params_tried"] = hyper_parameters_tried
            results["objective_values"] = objective_function_values
            results["surrogate_values"] = surrogate_values

            results["best_objective"] = np.max(results["objective_values"])
            results["best_params"] = results["params_tried"][
                np.argmax(results["objective_values"])
            ]
            if self.maximize_objective is False:
                results["objective_values"] = -1 * results["objective_values"]
                results["best_objective"] = -1 * results["best_objective"]

            if self.config.log_progress:
                tqdm.write("Best score: %s" % results["best_objective"])

            # check if early stop criteria has been met
            if self.config.early_stop(results):
                _logger.info("Early stopping criteria satisfied")
                break

        # saving the optimizer and ds in the tuner object which can save the surrogate function and ds details
        self.Optimizer = Optimizer

        return results

    def runRandomOptimizer(self):
        results = dict()

        X_sample_list = []
        Y_sample_list = []

        batch_size = self.config.batch_size
        n_iterations = self.config.num_iteration
        random_hyper_parameters = self.ds.get_random_sample(n_iterations * batch_size)

        # running the iterations
        iterator = (
            tqdm(range(0, len(random_hyper_parameters), batch_size))
            if self.config.log_progress
            else range(0, len(random_hyper_parameters), batch_size)
        )
        for idx in iterator:
            # getting batch by batch random values to try
            batch_hyper_parameters = random_hyper_parameters[idx : idx + batch_size]
            X_list, Y_list = self.runUserObjective(batch_hyper_parameters)

            X_sample_list = np.append(X_sample_list, X_list)
            Y_sample_list = np.append(Y_sample_list, Y_list)

            results["params_tried"] = X_sample_list
            results["objective_values"] = Y_sample_list

            results["best_objective"] = np.max(results["objective_values"])
            results["best_params"] = results["params_tried"][
                np.argmax(results["objective_values"])
            ]
            if self.maximize_objective is False:
                results["objective_values"] = -1 * results["objective_values"]
                results["best_objective"] = -1 * results["best_objective"]

            if self.config.log_progress:
                tqdm.write("Best score: %s" % results["best_objective"])

            # check if early stop criteria has been met
            if self.config.early_stop(results):
                _logger.info("Early stopping criteria satisfied")
                break

        return results

    def runUserObjective(self, X_next_PS):
        # initially assuming entire X_next_PS is evaluated and returned results are only Y values
        X_list_evaluated = X_next_PS
        results = self.objective_function(X_next_PS)
        Y_list_evaluated = results

        # if result is a tuple, then there is possibility that partial values are evaluated
        if isinstance(results, tuple):
            X_list_evaluated, Y_list_evaluated = results

        X_list_evaluated = np.array(X_list_evaluated)
        if self.maximize_objective is False:
            Y_list_evaluated = -1 * np.array(Y_list_evaluated)
        else:
            Y_list_evaluated = np.array(Y_list_evaluated)

        return X_list_evaluated, Y_list_evaluated



import warnings

import numpy as np
from sklearn.utils import check_random_state
from sklearn.utils.random import sample_without_replacement
from sklearn.model_selection import ParameterGrid


def parameter_sampler(
    param_dist: dict, n_size: int, *, random_state=None
) -> list[dict]:
    """
    :param param_dist: dict of parameter distributions
    :param n_size: number of samples to return
    :param random_state: specify to reproduce results
    :return: list of samples mapping each parameter to one of its
        allowed values.
    """
    rng = check_random_state(random_state)

    # if all distributions are given as lists, we want to sample without
    # replacement
    if all(not hasattr(v, "rvs") for v in param_dist.values()):
        # look up sampled parameter settings in parameter grid
        param_grid = ParameterGrid(param_dist)
        grid_size = len(param_grid)

        if grid_size < n_size:
            warnings.warn(
                "The total space of parameters %d is smaller "
                "than n_iter=%d. Running %d iterations. For exhaustive "
                "searches, use GridSearchCV." % (grid_size, n_size, grid_size),
                UserWarning,
            )
            n_size = grid_size
        return [
            param_grid[i]
            for i in sample_without_replacement(grid_size, n_size, random_state=rng)
        ]

    else:
        # Always sort the keys of a dictionary, for reproducibility
        items = sorted(param_dist.items())
        samples = []
        keys = []
        for k, v in items:
            keys.append(k)
            if hasattr(v, "rvs"):
                samples.append(v.rvs(random_state=rng, size=n_size).tolist())
            else:
                # forcing object dtype to avoid conversion of mixed datatypes
                v = np.array(v, dtype=object)
                samples.append(rng.choice(v, size=n_size))

        return [dict(zip(keys, sample)) for sample in zip(*samples)]



import functools


def serial(func):
    func.is_wrapped = True

    @functools.wraps(func)
    def wrapper(params_batch):
        return [func(**params) for params in params_batch]

    return wrapper


def parallel(n_jobs):
    from joblib import Parallel, delayed, cpu_count

    assert n_jobs != 0
    assert n_jobs >= -1

    def decorator(func):
        func.is_wrapped = True

        if n_jobs == -1:
            func.batch_size = cpu_count()
        else:
            func.batch_size = n_jobs

        @functools.wraps(func)
        def wrapper(params_batch):
            return Parallel(n_jobs)(delayed(func)(**params) for params in params_batch)

        return wrapper

    return decorator


def celery(n_jobs, timeout=None):
    from celery import exceptions

    def decorator(func):
        func.is_wrapped = True
        func.batch_size = n_jobs

        @functools.wraps(func)
        def wrapper(params_batch):
            async_results = [func(**params) for params in params_batch]
            results = []
            params_evaluated = []
            for params, async_result in zip(params_batch, async_results):
                try:
                    result = async_result.get(timeout=timeout)
                    params_evaluated.append(params)
                    results.append(result)
                except exceptions.TimeoutError:
                    # ignore the timed out jobs as we are returning (params, result) tuple
                    continue

            return params_evaluated, results

        return wrapper

    return decorator


def custom(n_jobs):
    assert n_jobs > 0

    def decorator(func):
        func.is_wrapped = True
        func.batch_size = n_jobs

        @functools.wraps(func)
        def wrapper(params_batch):
            return func(params_batch)

        return wrapper

    return decorator



from collections.abc import Mapping, Iterable
import math
import numpy as np
from collections.abc import Callable
from itertools import compress
from functools import cached_property
import warnings

from scipy.stats._distn_infrastructure import rv_frozen
from scipy.stats._multivariate import multi_rv_frozen
from sklearn.preprocessing import MinMaxScaler

from mango.domain.parameter_sampler import parameter_sampler


class DomainSpace:
    def __init__(
        self,
        param_dist: dict,
        *,
        param_sampler: Callable = parameter_sampler,
        constraint: Callable = None,
        constraint_max_retries: int = 10,
        scale_params: bool = False,
    ):
        self.dist = param_dist
        if not isinstance(self.dist, (Mapping)):
            raise TypeError(
                "Parameter distribution is not a dict or " "({!r})".format(self.dist)
            )

        for key in self.dist:
            if not isinstance(self.dist[key], Iterable) and not hasattr(
                self.dist[key], "rvs"
            ):
                raise TypeError(
                    "Parameter value is not iterable "
                    "or distribution (key={!r}, value={!r})".format(key, self.dist[key])
                )

        self.param_sampler = param_sampler
        self.constraint = constraint
        self.constraint_max_tries = constraint_max_retries
        self.scale_params = scale_params
        (
            self.categorical_params,
            self.integer_params,
            self.multivariate_params,
            self.univariate_params,
        ) = self.classify_parameters(self.dist)
        self.domain_size = self.calc_domain_size()

    @cached_property
    def categorical_index_lookup(self):
        res = {}
        for param in self.categorical_params:
            res[param] = {ele: pos for pos, ele in enumerate(self.dist[param])}
        return res

    @cached_property
    def multivariate_param_dimensions(self):
        res = {}
        for param in self.multivariate_params:
            res[param] = len(self.dist[param].mean())
        return res

    @cached_property
    def gp_scaler(self):
        data = self.get_domain()
        x_gp = self._convert_GP_space(data)
        scaler = MinMaxScaler()
        scaler.fit(x_gp)
        return scaler

    def get_domain(self):
        return self.get_random_sample(self.domain_size)

    def get_random_sample(self, size):
        if self.constraint is None:
            return self._get_random_sample(size)

        samples = []
        n_tries = 0
        factor = 2.0
        while len(samples) < size and n_tries < self.constraint_max_tries:
            n_samples = math.ceil((size - len(samples)) * factor)
            raw_samples = self._get_random_sample(n_samples)
            _filters = self.constraint(raw_samples)
            filtered_samples = list(compress(raw_samples, _filters))
            if len(filtered_samples) == 0:
                factor *= 10
            else:
                factor = max(len(raw_samples) / len(filtered_samples), 1.0)
            samples += filtered_samples
            n_tries += 1
        if len(samples) < size:
            warnings.warn(
                f"Could not get {size} samples that satisfy the constraint"
                f"even after {n_tries * size} random samples ",
                UserWarning,
            )
        return samples[:size]

    def _get_random_sample(self, size):
        return self.param_sampler(self.dist, size)

    def _convert_GP_space(self, domain_list):
        categorical_params = self.categorical_params

        X = []
        for domain in domain_list:
            curr_x = []
            for x in sorted(domain.keys()):
                # this is a categorical variable which require special handling
                if x in categorical_params:
                    size = len(self.dist[x])  # total number of categories.
                    # we need to see the index where domain[x] appears in param_dict[x]
                    index = self.categorical_index_lookup[x][domain[x]]
                    listofzeros = [0.0] * size
                    # We will set the value to one for one hot encoding
                    listofzeros[index] = 1.0
                    # expanding current list
                    curr_x = curr_x + listofzeros
                elif x in self.multivariate_params:
                    curr_x = curr_x + list(domain[x])

                # this value can be directly used, for int too, we will consider it as a float for GP
                else:
                    curr_x.append(domain[x])

            X.append(curr_x)

        X = np.array(X)
        return X

    def convert_GP_space(self, domain_list):
        """
        convert the hyperparameters from the param_dict space to the GP space, by converting the
        categorical variables to one hotencoded, and return a numpy array which can be used to train the GP

        input is the domain_list generated using the Parameter Sampler.
        """
        X = self._convert_GP_space(domain_list)
        if self.scale_params:
            X = self.gp_scaler.transform(X)
        return X

    def convert_PS_space(self, X_gp):
        """
        Convert from the X_gp space which is a numpy array that can be given input to the gaussian process
        to the parameter sampler space which is a list of the dict

        We have to reverse the one-hotencoded transformation of the categories to the category name
        """
        X_ps = []

        categorical_params = self.categorical_params
        int_params = self.integer_params
        mv_params = self.multivariate_params
        param_dict = self.dist

        if self.scale_params:
            X_gp = self.gp_scaler.inverse_transform(X_gp)

        for i in range(X_gp.shape[0]):

            curr_x_gp = X_gp[i]
            # we will create a list of dict, same as parameter samplers
            curr_x_ps = dict()
            index = 0

            # every sample is from the param_dict
            for par in sorted(param_dict.keys()):
                # this has to have integer values
                if par in int_params:
                    curr_x_ps[par] = int(round(curr_x_gp[index]))
                    index = index + 1

                # this par is a categorical variable and we need to handle it carefully
                elif par in categorical_params:
                    size = len(param_dict[par])  # total number of categories.
                    one_hot_encoded = curr_x_gp[index : index + size]
                    category_idx = np.argmax(one_hot_encoded)
                    category_val = param_dict[par][category_idx]
                    curr_x_ps[par] = category_val
                    # we have processed the entire one-hotencoded
                    index = index + size

                elif par in mv_params:
                    dim = self.multivariate_param_dimensions[par]
                    curr_x_ps[par] = curr_x_gp[index : index + dim].tolist()
                    index += dim

                # this is a float value
                else:
                    curr_x_ps[par] = curr_x_gp[index]
                    index = index + 1

            X_ps.append(curr_x_ps)
        return X_ps

    def calc_domain_size(self) -> int:
        """
        Return an estimate of number of points in the search space.
        """
        # Minimum and maximum domain size
        domain_min = 50000
        domain_max = 500000

        ans = 1

        for par, dist in self.dist.items():
            if par in self.univariate_params:
                loc, scale = dist.args
                min_scale = 1
                scale = int(scale)
                if scale < min_scale:
                    scale = min_scale

                ans = ans * scale * 50

            elif par in self.multivariate_params:
                # FIXME:  assuming scale 10 for multivariate as don't see a way to get it from the dist
                ans = ans * 10

            elif par in self.integer_params:
                ans = ans * len(dist)

            elif par in self.categorical_params:
                ans = ans * len(dist)

        if ans < domain_min:
            ans = domain_min

        if ans > domain_max:
            ans = domain_max

        return ans

    @staticmethod
    def classify_parameters(param_dict: dict) -> (set, set, set, set):
        """
        Identify parameters that are categorical or integer types.
        Categorical values/discrete values are considered from the list of each value being str
        Integer values are considered from list of each value as int or from a range
        Multivariate values are random variables defined using scipy stats multivariate distribution
        Univariate values are random variable defined using scipy stats distribution

        :param param_dict: dictionary of parameter distributions
        :return: a tuple of four sets where first element is the set of parameter that are categorical,
        second element is the set of parameters that are integer, third element is the set of params that are
        multivariate, and fourth element is the set of params that are univariate
        """
        categorical_params = set()
        int_params = set()
        multivariate_params = set()
        univariate_params = set()

        for par in param_dict:
            if isinstance(param_dict[par], rv_frozen):
                # FIXME: what if the distribution generators ints , GP would convert it to float
                univariate_params.add(par)
            elif isinstance(param_dict[par], multi_rv_frozen):
                multivariate_params.add(par)

            elif isinstance(param_dict[par], range):
                int_params.add(par)

            elif isinstance(param_dict[par], Iterable):
                # for list with all int we are considering it as non-categorical
                try:
                    # this check takes care of numpy ints as well
                    all_int = all(
                        x == int(x) and not isinstance(x, bool) for x in param_dict[par]
                    )
                except (ValueError, TypeError):
                    all_int = False

                if all_int:
                    int_params.add(par)
                # For lists with mixed type, floats or strings we consider them categorical or discrete
                else:
                    categorical_params.add(par)
            else:
                warnings.warn(f"Parameter {par} could not be classified")

        return (categorical_params, int_params, multivariate_params, univariate_params)
